#!/usr/bin/env python

"""
Simple indexing of sorted files.

Copyright (C) 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.

Motivations
-----------

When searching files, the cost to scan an entire file typically exceeds the
cost to navigate to the appropriate region and to scan a smaller portion of the
file, especially for large files. At the same time, exotic data structures
encouraging multiple seeks and reads are likely to waste time compared to just
performing a single read operation, even if that operation involves a larger
quantity of data, at least for storage with hard disk access characteristics.

Potential Improvements
----------------------

Ideally, the acquisition of records should be done more generally than just
reading lines, and the selection of matches should involve more than just
selecting the first column.
"""

import bisect

def index_by_lines(f, interval):

    """
    Index a file 'f', creating an index entry for a line after a given number,
    defined by 'interval', has been read since the last entry.
    """

    l = []
    pos = 0

    for i, line in enumerate(f.xreadlines()):
        columns = line.split("\t")
        if i % interval == 0:
            l.append((columns[0], pos))
        pos += len(line)

    return l

def find_line_with_index(f, l, term):

    """
    Find in file 'f', using the given index list 'l', the given 'term',
    returning a line employing the term or None if no such line was found.
    """

    i = bisect.bisect_left(l, (term, None))
    found, pos = l[i]

    # Since the index is more coarse than the underlying file, the bisect left
    # operation will most likely point to an index entry for later records than
    # the desired one.

    if found > term:
        i = max(0, i - 1)
        found, pos = l[i]

    f.seek(pos)
    return find_line_in_file(f, term)

def find_line_in_file(f, term):

    """
    Find in file 'f' the given 'term', returning a line employing the term or
    None if no such line was found.
    """

    for line in f.xreadlines():
        columns = line.split("\t")
        if term == columns[0]:
            return line

    return None

class Index:

    "An index abstraction."

    def __init__(self, entries, f):
        self.entries = entries
        self.f = f

    def find(self, term):
        return find_line_with_index(self.f, self.entries, term)

# vim: tabstop=4 expandtab shiftwidth=4
